
module Phase3 where
import Prelude          as P
import Dictionnaire     as D
import Data.Map.Strict  as M
import Data.Set         as S
import Autorise

type Nom = String
type Transitions = Dictionnaire Char Nom
data Etat = Etat { getFinal :: Bool
                 , getTrans :: Transitions
                 }
            deriving Show
data AutomateFini = AutomateFini { getEtats   :: (Map Nom Etat)
                                 , getInitial :: Nom
                                 }
                     deriving Show





-- Reconnaissance d'un mot par un automate
reconnise :: AutomateFini -> String -> Bool
reconnise (AutomateFini etats nom) s = reconniseE (etats ! nom) s  
     where
       reconniseE e ""         = getFinal e
       reconniseE e (c:w)      = reconniseM (rechercherParCle (getTrans e) c) w
       reconniseM Nothing w    = False
       reconniseM (Just nom) w = reconniseE (etats ! nom) w

-- créer un automate a partir d'une liste d'états avec leur booleen d'acceptation et d'une liste de transitions
-- l'état initial est le premier nom de la liste de noms
newAutomaton :: [(Nom,Bool)] -> [(Nom,Char,Nom)] -> AutomateFini
newAutomaton noms tr = AutomateFini (creerEtats noms tr) (fst (head noms))

creerEtats :: [(Nom,Bool)] -> [(Nom,Char,Nom)] -> Map Nom Etat
creerEtats noms tr = mapWithKey (creerEtat tr) (M.fromList noms)

creerEtat :: [(Nom,Char,Nom)] -> Nom -> Bool -> Etat
creerEtat tr e b = Etat b (insererListe creer transitions)
   where
     tr_e = P.filter (\(e',_,_)->e==e') tr
     transitions = fmap (\(_,c,n)->(c,n)) tr_e

-- type des automate infinis du TP2
data EtatD = EtatD Bool (Char->EtatD)
type AutomateD = EtatD

-- projeter sur le type du TP2
forgetFiniteness :: AutomateFini -> AutomateD
forgetFiniteness (AutomateFini etats init) = forgetFinitenessEtat (etats ! init)
   where
   forgetFinitenessEtat (Etat b tr) =
       EtatD b (forgetFinitenessTransition tr)
   forgetFinitenessTransition tr c =
       let nom = recherche tr c
       in let etat = etats ! nom
       in forgetFinitenessEtat etat


-- projeter sur le type du TP2
forgetFinitenessSafe :: AutomateFini -> AutomateD
forgetFinitenessSafe (AutomateFini etats init) = forgetFinitenessEtat (etats ! init)
   where
   forgetFinitenessEtat (Etat b tr) =
       EtatD b (enleverMaybe.(forgetFinitenessTransition tr))
   forgetFinitenessTransition tr c = do
       nom  <- rechercherParCle tr c
       etat <- etats !? nom
       return (forgetFinitenessEtat etat)
   enleverMaybe (Just e) = e
   enleverMaybe Nothing = EtatD False (\c-> enleverMaybe Nothing)



-- version non-déterministe
type NomND = String
type TransND = Dictionnaire Char [NomND]
data EtatND = EtatND { getFinalND :: Bool
                     , getTransND :: TransND
                     }
                     deriving Show
data AutomateFiniND = AutomateFiniND { getEtatsND :: Map NomND EtatND
                                     , getInitND   :: [NomND]
                                     }
                     deriving Show

-- fonction de determinisation
determinise :: AutomateFiniND -> AutomateFini
determinise (AutomateFiniND etatsND initND) = AutomateFini etats init
   where
   init  = nommer initND
   etats = determiniseAux etatsND M.empty [initND] 

determiniseAux :: (Map NomND EtatND) -> Map Nom Etat -> [[NomND]] -> Map Nom Etat
determiniseAux etatsND acc [] = acc
determiniseAux etatsND acc (noms:l)
        | M.member nomDet acc = determiniseAux etatsND acc l
	| otherwise           = determiniseAux etatsND (M.insert nomDet etat acc) ((D.toListVal ft)++l)
	where
	  nomDet   = nommer noms
	  etat     = Etat (estTerminalList es) (mapDico nommer ft)
          es       = fmap (etatsND !) noms
	  ft       = fusionT es


estTerminalList :: [EtatND] -> Bool
estTerminalList es = any getFinalND es
 

regroupe :: [TransND] -> TransND
regroupe = (mapDico concat) . unionCons 

fusionT :: [EtatND] -> TransND
fusionT = regroupe . (fmap getTransND)



nommer :: [NomND] -> Nom
nommer = show . S.fromList



-- créer un automate qui reconnaît un caractère compris dans un segment donné
segment :: Char -> Char -> AutomateFiniND
segment c1 c2 = AutomateFiniND (M.fromList [([c1,c2,'1'],etat1) , ([c1,c2,'2'],etat2) ])
                               [[c1,c2,'1']]
               where
	          etat1 = EtatND False (D.insererIntervalle D.creer c1 c2 [[c1,c2,'2']])
	          etat2 = EtatND True D.creer


-- créer un automate qui reconnaît la somme des langages reconnus par les automates entres
(!+!) :: AutomateFiniND -> AutomateFiniND -> AutomateFiniND
e1 !+! e2 =
     AutomateFiniND (M.union etats1 etats2) (i1 ++ i2)
     where
      (AutomateFiniND etats1 i1) = modifNom e1 '1'
      (AutomateFiniND etats2 i2) = modifNom e2 '2'
     

modifNom :: AutomateFiniND -> Char -> AutomateFiniND
modifNom (AutomateFiniND etats i) c = AutomateFiniND (mapKeys (c:) (fmap (modifNomE c) etats)) (fmap (c:) i)
modifNomE :: Char -> EtatND -> EtatND
modifNomE c (EtatND b tr) = EtatND b (mapDico (fmap (c:)) tr)

-- créer un automate qui reconnaît l'étoile du langage reconnu par l'automate entre
star :: AutomateFiniND -> AutomateFiniND
star e = AutomateFiniND (M.insert "0" etat0 (fmap cont etats)) ["0"]
     where
       cont (EtatND False tr) = EtatND False tr
       cont (EtatND True tr) = EtatND True (regroupe [tr,transInit])
       AutomateFiniND etats i = modifNom e '1'
       transInit = regroupe (fmap (getTransND . (etats !)) i)
       etat0 = EtatND True transInit


-- créer un automate qui reconnaît la concaténation des langages reconnus par les automates entres
(!.!) :: AutomateFiniND -> AutomateFiniND -> AutomateFiniND
e1 !.! e2 =
     AutomateFiniND (M.union (fmap cont etats1) etats2) i1
     where
       cont (EtatND False tr) = EtatND False tr
       cont (EtatND True  tr) = EtatND (estTerminalList  (fmap (etats2!) i2))
                                       (regroupe (tr:(fmap (getTransND . (etats2 !)) i2 )))
       AutomateFiniND etats1 i1 = modifNom e1 '1'
       AutomateFiniND etats2 i2 = modifNom e2 '2'


-- type des expressions régulières
data Regexp = Segment Char Char  |  Regexp :+: Regexp  |  Star Regexp  |  Regexp :.: Regexp |  Eps


-- construire un automate reconnaissant le même langage que l'expression régulière entrée
regexp2Automata :: Regexp -> AutomateFini
regexp2Automata = determinise . regexp2AutomataND

regexp2AutomataND :: Regexp -> AutomateFiniND
regexp2AutomataND (Segment c1 c2) = segment c1 c2
regexp2AutomataND (r1 :+: r2) = regexp2AutomataND r1 !+! regexp2AutomataND r2
regexp2AutomataND (r1 :.: r2) = regexp2AutomataND r1 !.! regexp2AutomataND r2
regexp2AutomataND (Star r) = star (regexp2AutomataND r)
regexp2AutomataND Eps =  AutomateFiniND (M.fromList [("",EtatND True D.creer)]) [""]




-- minimiser un automate (bonus)
minimisation :: AutomateFini -> AutomateFini
minimisation e@(AutomateFini etats i) = AutomateFini etatsMin ((transformation e) ! i)
  where
  etatsMin = foldlWithKey (\acc n1 n2 -> M.insert n2 (transfEtat (etats ! n1)) acc) M.empty (transformation e)
  transfEtat (Etat b tr) = Etat b (mapDico ((transformation e) !) tr)


transformation :: AutomateFini -> Map Nom Nom
transformation  (AutomateFini etats i) = minAux (size init) init
  where
  init = fmap (show . getFinal) etats
  size = S.size . S.fromList . M.elems
  contr m (Etat b tr) = show b ++ D.imprimer (mapDico (m !) tr)  
  minAux s acc
      | s == s'   = acc
      | otherwise = minAux s' res
      where
        res = fmap (contr acc) etats  
        s'  = size res
	       
